package tools.descartes.dql.connector.kieker;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;

import org.eclipse.emf.ecore.util.EcoreUtil;

import edu.kit.ipd.descartes.solving.probes.AvgThroughputProbe;
import kieker.common.configuration.Configuration;
import tools.descartes.dql.connector.AbstractOSGiQueryConnector;
import tools.descartes.dql.connector.PerformanceMetricsQueryConnector;
import tools.descartes.dql.connector.QueryProgressListener;
import tools.descartes.dql.connector.kieker.configgenerator.XMLGenerationManager;
import tools.descartes.dql.connector.kieker.filter.KiekerController;
import tools.descartes.dql.connector.kieker.structures.Resource;
import tools.descartes.dql.connector.kieker.structures.Trace;
import tools.descartes.dql.connector.kieker.structures.identifier.ResourceIdentifier;
import tools.descartes.dql.connector.kieker.structures.records.ResourceRecord;
import tools.descartes.dql.models.mapping.domain.ContinuousSeriesResult;
import tools.descartes.dql.models.mapping.domain.DecimalResult;
import tools.descartes.dql.models.mapping.domain.DiscreteSeriesResult;
import tools.descartes.dql.models.mapping.domain.DomainFactory;
import tools.descartes.dql.models.mapping.domain.Entity;
import tools.descartes.dql.models.mapping.domain.Probe;
import tools.descartes.dql.models.mapping.domain.Result;
import tools.descartes.dql.models.mapping.domain.ResultList;
import tools.descartes.dql.models.mapping.domain.SeriesResultElement;
import tools.descartes.dql.models.mapping.domain.StatType;
import tools.descartes.dql.models.mapping.mapping.EntityMapping;

public class PerformanceMetricsQueryConnectorImpl extends
		AbstractOSGiQueryConnector implements PerformanceMetricsQueryConnector {

	private static String MAX_TIMEOUT_KEY = "kieker.maxRunTimeInSeconds";
	private static String KIEKER_CONFIG_KEY = "kieker.monitoring.default.properties";
	private static String APPLICATION_JAR_FILE = "kieker.applicationJarFile";
	private static String KIEKER_JAR_FILE = "kieker.kiekerJarFile";
	private static String KIEKER_LOG_DIRECTORY = "kieker.log.directory";

	private final DomainFactory domainFactory = DomainFactory.eINSTANCE;


	@Override
	public EntityMapping requestMetrics(final EntityMapping request) {
		System.out.println("Starting Kieker Controller now" + request.toString());

		final EntityMapping response = EcoreUtil.copy(request);
		final String propertiesLocation = request.getModelLocation();

		// Properties laden
		Properties prop = getProperties(propertiesLocation);
		String projectPath = propertiesLocation.substring(0,
				propertiesLocation.lastIndexOf(File.separator) + File.separator.length());


		String appLocation = (String) prop.get(APPLICATION_JAR_FILE);
		String kiekerLocation = (String) prop.get(KIEKER_JAR_FILE);
		String logDirectory = (String) prop.get(KIEKER_LOG_DIRECTORY);

		ArrayList<String> services = KiekerHelper.getServiceIdentifiers(request);

/*
		// Config Generation
		if (!XMLGenerationManager.generateDefaultXML(services, projectPath+File.separator+"META-INF")) {
			System.out.println("Warning: Config could not be created");
		}

		//TODO implement wait for end
		// TODO implement warning if no int
		int timeOut = Integer.valueOf(prop.getProperty(MAX_TIMEOUT_KEY));
		String kiekerConfigLocation = prop.getProperty(KIEKER_CONFIG_KEY);
		Properties kiekerProperties = loadKiekerProperties(kiekerConfigLocation);
		//String newOutputPath = getNewOutPutPath(projectPath);
		String newOutputPath = projectPath+File.separator+"staticLog";
		kiekerProperties.setProperty("kieker.monitoring.writer.filesystem.AsyncFsWriter.customStoragePath",
				newOutputPath);
		try {
			kiekerProperties.store(new FileOutputStream(kiekerConfigLocation), "");
		} catch (FileNotFoundException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		} catch (IOException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
		
			
		// Kieker starten
		// TODO kieker.monitoring.default.properties ver?ndern/erstellen
		Process p;
		ResourceMonitor.startResourceMonitor("D:\\Master\\GR\\kieker-1.13-SNAPSHOT","D:\\Master\\GR\\Test");;
		try {
			//monitor = new ResourceMonitor();
			//monitor.startMonitoring(2, 5);
			
			String command = "java -javaagent:" + kiekerLocation + " -jar " + appLocation;
			p = Runtime.getRuntime()
					.exec(command);
			String line;
			BufferedReader in = new BufferedReader(new InputStreamReader(p.getInputStream()));
			while ((line = in.readLine()) != null) {
				System.out.println(line);
			}
			in.close();
		} catch (Exception e) {
			System.out.println(e.getMessage());
			e.printStackTrace();
			throw new IllegalStateException("Could not execute application with kieker");
		}

		// warten bis timeout
		try {
			Thread.sleep(timeOut * 1000);
			//TODO run directly to avoid having to kill java.exe
			// Runtime.getRuntime().exec("taskkill /f /im java.exe");
			// Runtime.getRuntime().exec("taskkill /f /im cmd.exe");
			p.destroy();
			//System.out.println("Process destroyed: " + p.toString());
			//monitor.stopMonitoring();
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
*/
		// Kieker ergebnisse einlesen
		// override log path for tests with static folders
		String newOutputPath = logDirectory;
		 String[] traceFolders = KiekerHelper.getTraceFolders(newOutputPath);
		 Configuration readerConfig = new Configuration();
		 
		 for (int i = 0; i< traceFolders.length; i++){
			 readerConfig.setProperty("TRACE_"+i, newOutputPath+File.separator+traceFolders[i]);
		 }
		 
		 for (tools.descartes.dql.models.mapping.domain.Resource r : response.getResources()) {
			  for (Probe prob : r.getProbes()) {
				  String r_id = r.getIdentifier();
				  String r_type = KiekerHelper.getKiekerMetric(prob.getMetricName());
				  String r_value = KiekerHelper.getHostName()+" & SYSTEM 0.0 USER 0.0 "+r_type+" 0";
				  //String r_value = KiekerHelper.getHostName();
				  readerConfig.setProperty("RESOURCE_"+r_id, r_value);
			  }
		  }
		 
		 


		
			
			
		
		 

		// Start KiekerController with custom config (currently only location of
		// traceFiles)
		 KiekerController c = new KiekerController(readerConfig);

		// Fill the traces according to the applied filters
		 c.fillTrace();

		// Get the result trace
		 Trace result = c.getTrace();

		// Process the trace back to a DQL result
		// Kieker ergebnisse filtern
			
		 
		 processKiekerSolverResults(response, result);
		return response;
	}

	private String getNewOutPutPath(String standardBatPath) {
		standardBatPath += File.separator + "logs";
		File logFolder = new File(standardBatPath);
		if (!logFolder.exists()) {
			logFolder.mkdir();
			logFolder.mkdirs();
		}
		int i = 1;
		File subLogFolder = null;
		do {
			subLogFolder = new File(logFolder, "run" + i);
			i++;
		} while (subLogFolder.exists());
		subLogFolder.mkdir();
		subLogFolder.mkdirs();
		return subLogFolder.getAbsolutePath();
	}

	/* Modifies response to include the results of Kieker */
	private void processKiekerSolverResults(EntityMapping response, Trace kiekerResults) {
		
		
		// Process Ressources
		//if (kiekerResults.getResourceIdentifierToResourceMap() != null){

		// Map containing the resulting records sorted by resources
		//final HashMap<String, List<ResourceRecord>> kiekerResultMap = new HashMap<String, List<ResourceRecord>>(kiekerResults.getResourceIdentifierToResourceMap().size());


		// kieker resource id to resource mapping, help for identifying the
		// records
		//Map<ResourceIdentifier, Resource> resourceMap = kiekerResults.getResourceIdentifierToResourceMap();

		// for every retrieved record get the original dql identifier, if no
		// match found but still in map, throw away)
		// TODO: should be avoided by resource filter
		/*for (ResourceRecord kiekerProbe : records) {
			
			//if (resourceMap.get(kiekerProbe.getIdentifier()) != null) {
				//String identifier = resourceMap.get(kiekerProbe.getIdentifier()).getName();
				String identifier = "r1";
				List<ResourceRecord> kiekerProbes = kiekerResultMap.get(identifier);
				if (null == kiekerProbes) {
					kiekerProbes = new ArrayList<ResourceRecord>();
					kiekerProbes.add(kiekerProbe);
					kiekerResultMap.put(identifier, kiekerProbes);

				} else {
					kiekerProbes.add(kiekerProbe);
					kiekerResultMap.put(identifier, kiekerProbes);

				}
			

		}*/
		
		// List of Records for one resource, value for kiekerResultMap
		
		List<ResourceRecord> records = kiekerResults.getResources();
		
		if (records != null && records.size()>0){
		long firstTs = records.get(0).getTimestamp();
		for (ResourceRecord r: records){
			if (r.getTimestamp() < firstTs){
				firstTs = r.getTimestamp();
			}
		}
		
		
		final ArrayList<Entity> entities = new ArrayList<Entity>();
		entities.addAll(response.getResources());
		entities.addAll(response.getServices());
		
		// browse through each entity and its probes 
		//   and try to replace the probe with a result
		for (Entity ent : entities) {
			for (Probe dqlProbe : ent.getProbes()) {
				for (StatType st : dqlProbe.getStatTypes()) { 
						if (KiekerHelper.METRIC_UTILIZATION.equalsIgnoreCase(dqlProbe.getMetricName())) {
							if (st.getTypeName().equals("DEFAULT") || st.getTypeName().equals("SAMPLE")){
								//List<ResourceRecord> kiekerProbes = kiekerResultMap.get( ent.getIdentifier().toString());
								ContinuousSeriesResult res = domainFactory.createContinuousSeriesResult();
								if (records != null) {
									try {
										for (ResourceRecord kiekerProbe : records) {
												
													SeriesResultElement e = domainFactory.createSeriesResultElement();
													// convert ts to millis
													BigDecimal time = new BigDecimal((kiekerProbe.getTimestamp()-firstTs) / 1000000.0);
													BigDecimal val = new BigDecimal(kiekerProbe.getContent().get(2));
													System.out.println("TIME: "+time+" / UTIL: "+val);
													e.setX(time);
													e.setY(val);
													res.getElements().add(e);
													res.setValid(true);
										}
										res.setValid(true);
									} catch (NumberFormatException exc) {
										res.setValid(false);
									}
									st.setResult(res);
									
								}
								
						}  else {
							// do nothing
						}
					}
					
				}
			}
		}
		}
	}

	private Properties loadKiekerProperties(final String propertiesLocation) {
		final Properties properties = new Properties();
		File file = new File(propertiesLocation);

		if (file == null || !file.exists()) {
			throw new IllegalStateException(
					"Could not find referenced .properties file");
		}

		try {
			properties.load(new FileInputStream(file));
		} catch (final IOException e) {
			throw new IllegalStateException(
					"Could not load properties instance");
		}
		return properties;
	}

	private Properties getProperties(final String propertiesLocation) {
		final Properties properties = new Properties();
		File file = new File(propertiesLocation);

		if (file == null || !file.exists()) {
			throw new IllegalStateException(
					"Specified file in USING not accessible, unable to execute the query further. Aborting.");
		}

		try {
			properties.load(new FileInputStream(file));
		} catch (final IOException e) {
			throw new IllegalStateException(
					"Could not load properties instance from specified model location in USING clause");
		}
		// Add absolutePath for the bat file
		String propertiesFolder = new File(propertiesLocation).getParentFile().getAbsolutePath();

		properties.setProperty(APPLICATION_JAR_FILE,
				propertiesFolder + File.separator + properties.get(APPLICATION_JAR_FILE));
		properties.setProperty(KIEKER_JAR_FILE,
				propertiesFolder + File.separator + properties.get(KIEKER_JAR_FILE));
		properties.setProperty(KIEKER_CONFIG_KEY,
				propertiesFolder + File.separator + properties.get(KIEKER_CONFIG_KEY));
		properties.setProperty(KIEKER_LOG_DIRECTORY,
				propertiesFolder + File.separator + properties.get(KIEKER_LOG_DIRECTORY));
		return properties;
	}

	@Override
	public void reset() {
		// Nothing to do here
	}

	@Override
	public void addQueryProgressListener(QueryProgressListener arg0) {
		// TODO Auto-generated method stub

	}

	@Override
	public String getConstraint() {
		return null;
	}

	@Override
	public String getDoFConstraint() {
		return null;
	}

	@Override
	public String getDoFExplorationStrategy() {
		return null;
	}

	@Override
	public Properties getDoFExplorationStrategyProperties() {
		return null;
	}

	@Override
	public List<String> getSupportedConstraintNames() {
		return null;
	}

	@Override
	public List<String> getSupportedDoFConstraintNames() {
		return null;
	}

	@Override
	public List<String> getSupportedExplorationStrategyNames() {
		return null;
	}

	@Override
	public List<String> getSupportedTimeUnits() {
		return null;
	}

	@Override
	public List<EntityMapping> requestDoFMetrics(final EntityMapping request) {
		return null;
	}
	
	@Override
	public void setConstraint(final String constraint) {
	}

	@Override
	public void setDoFConstraint(final String doFConstraint) {
	}

	@Override
	public void setDoFExplorationStrategy(final String strategy,
			final Properties properties) {
	}

	@Override
	public void setObservationPeriod(final Date start, final Date end,
			final long sampleSize) {
	}

	@Override
	public void setObservationPeriod(final long amount, final String unit,
			final long sampleSize) {
	}

	@Override
	public void setObservationReference(final String type,
			final String identifier, final long sampleSize) {
	}

	@Override
	public boolean supportsConstraintRequests() {
		return false;
	}

	@Override
	public boolean supportsDoFRequests() {
		return false;
	}

	@Override
	public boolean supportsMetricRequests() {
		return true;
	}

	@Override
	public boolean supportsObservationPeriods() {
		return false;
	}
}